# FILE: index.tcl
#
# The [package index] command of the package "package".
#
# Usage:
# 	package index $package $directory $file
#
# Creates a pkgIndex.tcl file for the package $package which is
# [package provide]d by sourcing $file, and places that pkgIndex.tcl
# file in $directory, overwriting any previous pkgIndex.tcl file which
# may already be there.  $file should be specified as a relative
# pathname, relative to $directory.
#
# [package index] performs the same conceptual task as the built-in
# command [pkg_mkIndex] from the Tcl library -- creating a pkgIndex.tcl
# file.  However, [package index] is not a drop-in replacement for
# [pkg_mkIndex].  [package index] imposes more constrants on its
# pkgIndex.tcl files, which makes the task of generating those files
# considerably less complex, while at the same time generating index
# files which permit more robust handling of package dependencies.
#
# [package index] creates pkgIndex.tcl files which use
# [package prerequisites] to declare the dependencies of a package and
# which use [package present] to guarantee the prerequisites of a
# package are satisfied before [package ifneeded] is invoked to make a
# package available for loading.  Because all of this logic is specific
# to one particular version of one particular package, pkgIndex.tcl
# files generated by [package index] have the constraint that they index
# exactly one version of exactly one package.  Since there is at most
# one pkgIndex.tcl file contained in any directory, obviously, it
# follows that packages indexed by [package index] will be installed
# one version of one package per directory.  In contrast, [pkg_mkIndex]
# will index multiple packages installed in a single directory, though
# I don't believe the practice is encouraged.
#
# Since each invocation of [package index] creates one pkgIndex.tcl file
# indexing one version of one package, it can be assumed that the caller
# knows what package is being indexed.  That package name is the first
# argument, and is used to check that the sourcing of $file actually
# provides a version of that package.
#
# Another limitation of [package index] compared to [pkg_mkIndex] is
# that the pkgIndex.tcl files generated by [package index] support only
# the "direct loading" style of packages.  If "lazy loading" of portions
# of a package is desired, that can still be achieved via direct use of
# the auto-loader by the package.  See this package itself as an
# example.  There is no need to support lazy loading directly in
# pkgIndex.tcl, and leaving it out greatly simplifies [package index].
#
# [package index] sources $file in a slave interpreter to determine
# which $version of $package is provided.  The slave interpreter uses a
# special implementation of [package] to record the set of package
# prerequisites of $package: $prereq1, $prereq2,...  Then the
# pkgIndex.tcl file is created which has the following structure:
# 
# if {[string match "" [package provide package]} {return}
# if {[catch {package present -exact package 0.0.1}]} {return}
# package prerequisites $package $version { $prereq1 $prereq2 ... }
# if {[catch {eval package present $prereq1}]} {return}
# if {[catch {eval package present $prereq2}]} {return}
# ...
# package ifneeded $package $version [list source [file join $dir $file]]
#
# The current implementation of [package index] only indexes Tcl-only
# files which can be sourced.  Future release of package are intended
# to support a [package index] command which can index binary packages.
#
# Normally, [package index] should be evaluated only by the installation
# procedure of a package, immediately after the files of a package are
# installed into their installation directories.

proc ::package::index {package directory file} {

    if {[string compare relative [file pathtype $file]]} {
	return -code error "filename is not relative path: $file"
    }
    if {![file writable $directory]} {
	return -code error "can't write file pkgIndex.tcl in $directory"
    }
    set absfile [file join $directory $file]
    if {![file readable $absfile]} {
	return -code error "can't read file: $absfile"
    }

    # Create slave in original working directory in case TCL_LIBRARY is
    # a relative path (e.g. in the test suite). 
    set c [interp create]

    set oldDir [pwd]
    if {[catch {cd $directory} msg]} {
	cd $oldDir
	interp delete $c
	return -code error \
		"can't change working directory to $directory"
    }

    # Load this package into the slave.
    variable library
    variable version
    $c eval [list lappend auto_path [file dirname $library]]
    $c eval [list package require -exact package $version]

    $c eval {
	# Create namespace in slave
	namespace eval ::package::index {
	    variable prerequisites [list]
	    variable version
	    variable soughtPackage
	    namespace export package
	}
    }

    $c eval [list set ::package::index::soughtPackage $package]

    $c eval {
	# Wrap [package].  Replacement [package] will register
	# package dependencies when $file is sourced in slave.
	proc ::package::index::package {args} {
	    variable prerequisites
	    set option [lindex $args 0]
	    switch -glob -- $option {
		in* {
		    if {[string first $option index] != 0} {
			eval [::package::Wrap ::package::index:: [list]]
		    }
		    # do nothing; recursive calls to [package index]
		    # are prevented.
		}
		m* {
		    if {[string first $option mkdepend] != 0} {
			eval [::package::Wrap ::package::index:: [list]]
		    }
		    if {[catch {
			    uplevel ::package::index:: $args
		    } msg]} {
			if {![string match \
				"can't satisfy requirement*" $msg]} {
			    return -code error $msg
			}
		    }
		    # The argument to [package mkdepend] is a list of
		    # the prerequisites of the package being indexed.
		    # Record them and return an empty list so the same
		    # requirements will not be passed to
		    # [package require].
		    foreach req [lindex $args 1] {
			lappend prerequisites $req
		    }
		    return [list]
		}
		pro* {
		    if {[string first $option provide] != 0} {
			eval [::package::Wrap ::package::index:: [list]]
		    }
		    if {[llength $args] != 3} {
			eval [::package::Wrap ::package::index:: [list]]
		    }
		    # [package provide $package $version] found in the
		    # package being indexed.  Check that it is the
		    # package we seek, and record the version.
		    variable soughtPackage
		    set package [lindex $args 1]
		    if {[string compare $soughtPackage $package]} {
			return -code error -errorcode [list package] \
				"[info script] provides $package, not\
				sought package $soughtPackage"
		    }
		    variable version
		    set version [lindex $args 2]
		    # Recording complete.  Invoke [package provide]
		    eval [::package::Wrap ::package::index:: [list]]
		}
		r* {
		    if {[string first $option require] != 0} {
			eval [::package::Wrap ::package::index:: [list]]
		    }
		    # record the arguments to [package require] as 
		    # prerequisites of the package being indexed.
		    # Do *not* actually invoke [package require].
		    lappend prerequisites [lrange $args 1 end]
		}
		default {
		    eval [::package::Wrap ::package::index:: [list]]
		}
	    }
	}
	rename package ::package::index::
	namespace import ::package::index::package

	# Disable auto-loading and errors from unknown commands
	proc unknown args {}
	proc auto_import args {}

    }
	
    if {[catch {$c eval [list source $file]} msg]} {
	global errorCode errorInfo
	cd $oldDir
	interp delete $c
	if {[string compare package [lindex $errorCode 0]]} {
	    return -code error -errorcode $errorCode \
		    -errorinfo $errorInfo $msg
	} else {
	    return -code error "can't index $file:\n  $msg"
	}
    }

    set v [$c eval set ::package::index::version]
    set prereqs [$c eval set ::package::index::prerequisites]
    interp delete $c

    set index [subst -novariables {# Tcl package index file, version 1.2
# This file was generated by the Tcl command:
#	 package index [lrange [info level 0] 1 end]
# The command "package index" was provided by package [package provide package].
# Use that package and that command to re-generate this file.
# Do not edit this file by hand unless you are sure you know what you
# are doing.
#
# This file is sourced either when an application starts up or by a
# "package unknown" script.  It invokes the "package ifneeded" command
# to set up package-related information so that packages will be loaded
# automatically in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the full path name
# of this file's directory.
}]
    append index [subst -nocommands {#
# This file indexes package $package $v

if {[string match "" [package provide package]]} {return}
if {[catch {package present -exact package $version}]} {return}
}]

    append index [list package prerequisites $package $v $prereqs]\n

    foreach req $prereqs {
	append index [format \
		{if {[catch {package present %s}]} {return}} $req]\n
    }

    append index [list package ifneeded $package $v]
    append index { [list source [file join $dir }
    append index [list $file]]]

    set f [open pkgIndex.tcl w]
    puts $f $index
    close $f
    cd $oldDir
}


