# FILE: mkdepend.tcl
#
# The [package mkdepend] command of the package "package".
#
# Usage:
#	package mkdepend $requirementList
#
# Takes a list of package interface requirements, $requirementList, and
# returns an expanded list of package interface requirements.  
#
# Each element of $requirementList is a list of arguments suitable for
# passing to [package require].  The expanded list is created from
# $requirementList by examining what packages have been
# [package provide]-d, what packages have been [package ifneeded], and
# what packages have prerequisites registered with
# [package prerequisites].  The returned list is ordered so that if each
# of its elements is passed as a list of arguments to [package require],
# all the [package require]s will succeed.  If such a list cannot be
# constructed, [package mkdepend] returns an error, and a message
# explaining why the requirements in $requirementList cannot be
#  satisfied.
#
# Typical usage is at the start of the initialization files of a package
# or application.  For example:
#
#	foreach req [package mkdepend {
#		{foo 1}
#	    }] {
#	    eval package require $req
#	}
#
# The idea is that this package or application has declared that it
# requires the package interface foo 1, and it is up to
# [package mkdepend] to determine what additional interface requirements
# that might imply, given the environment of installed packages.
#

proc ::package::mkdepend {requirementList} {

    # Confirm the argument is a valid Tcl list of valid Tcl lists
    if {[catch {llength $requirementList} msg]} {
	return -code error "bad requirementList: $msg"
    }
    foreach req $requirementList {
	if {[catch {llength $req} msg]} {
	    return -code error "bad requirement \"$req\" in\
		    requirementList: $msg"
	}
    }

    # Normalize the requirementList so that it is a unique string to use
    # as an index in the Alternatives array
    set normList [list]
    foreach req $requirementList {
	if {[string match -exact [lindex $req 0]]} {
	    if {[llength $req] != 3} {
		return -code "bad requirement \"$req\" in\
			requirementList: should be \"-exact package\
			version\""
	    }
	    # Convert -exact syntax to minVersion == maxVersion syntax
	    set req [lrange $req 1 end]
	    lappend req [lindex $req end]
	}

	# Strip trailing ".0" from version numbers so equivalent version
	# numbers map to the same string.
	foreach idx {1 2} {
	    if {[llength $req] > $idx} {
		set version [lindex $req $idx]
		while {[regsub {[.]0+$} $version {} version]} {}
		set req [lreplace $req $idx $idx $version]
	    }
	}

	# Force element to be a list --> normalizes whitespace
	lappend normList [lrange $req 0 end]
    }

    # Reset the Alternatives array.  See [package::Alternatives]
    # below for a description of what it contains.
    variable Alternatives
    catch {unset Alternatives}
    array set Alternatives [list [list] \
	    [list 0 [list [list [list] [list]] [list] ] ] ]

    # Reset the Satisfiers array.  See [package::Satisfiers]
    variable Satisfiers
    catch {unset Satisfiers}
    array set Satisfiers [list Tcl \
	    [list 0 [list [list Tcl [package provide Tcl]] [list]]]]

    if {[catch {Alternatives $normList} altList]} {
	return -code error $altList
    }

# NOTE: Uncomment the following lines for more information about the
# alternatives available that satisfy $requirementList.
#puts "Alternatives:"
#foreach alt $altList {
#    puts [TopologicalSort $alt]
#}
    # Return a topological sort of the first alternative
    TopologicalSort [lindex $altList 0]
}

#		*	*	*	*	*

# [Alternatives] takes a list of requirements and returns a list of
# alternatives which satisfy all the requirements, ordered by
# preference.
#
# An "alternative" is a data structure which represents one alternative
# way to satisfy a requirementList.  An alternative is made up of two
# major parts.  
#
# First is a directed acyclic graph (DAG) representing the poset of
# package names partially ordered by the prerequisite relationship.  An
# edge A -> B is in the graph if in this alternative, the package B has
# package A as a prerequisite.  Also, there is no package C such that A
# is a prerequisite for C and C is a prerequisite for B.  The DAG is
# implemented in terms of Tcl data structues as a Tcl array DAG.  An
# edge A -> B is in the graph iff B is an element of the list $DAG(A).
# In addition, a list of the maximal elements of the DAG is maintained
# (the list of all packages A for which there is no edge A -> X) to
# improve the efficiency of certain operations.
#
# The second major part of an alternative is a map from each package
# name to a pair.  The first element of the pair is the version of
# that package in this alternative.  The second element of the pair
# is the corresponding requirement which when passed as an argument
# to [package require] will cause that version of the package to be
# loaded, assuming all prerequisites are satisfied.  This map is
# implemented as a Tcl array.
#
# So, each alternative is a pair of a DAG and a map.  Each DAG is a pair
# of a Tcl array and a Tcl list.  Each map is a Tcl array.  When the
# value of an alternative is needed for passing around as an argument
# from one proc to another, all the arrays are converted to their list
# representation as given by [array get].  
#
# This proc calls itself recursively to determine the full requirements
# of packages which require packages which require packages which...
#
proc ::package::Alternatives {requirementList} {

    # [Alternatives] stores in the package variable "Alternatives" the
    # partial results of its work.  Each slot in the array is indexed by
    # a normalized requirementList which has been an argument in a call
    # to [Alternatives].  The value of each slot is a list.  The
    # first element in that list is an integer code.  The remainder of
    # the list is interpreted based on the value of the code as follows.
    #
    # Code == 0.  This means the full set of alternatives for that
    # requirement has already been computed.  Each remaining element in
    # the list is one of the alternatives that satisfies that
    # requirement.  
    #
    # Code == 1.  This means that it has already been determined that
    # the requirement cannot be satisfied.  The next (second) element in
    # the list is a suitable error message.
    #
    # Code == -1.  This means that an attempt to determine the
    # alternatives for this requirement has already begun, but has not
    # yet completed.  If another attempt to determine alternatives for
    # this requirement tries to start, it is an indication of a
    # circular package dependency which will lead to infinite recursion
    # in the Alternatives proc.

    variable Alternatives
    if {![info exists Alternatives($requirementList)]} {
	# Initialize with one alternative: the empty alternative
	set Alternatives($requirementList) \
	    [list -1 [list [list [list] [list]] [list] ] ]

	variable Prerequisites
	set prior [list]
	foreach requirement $requirementList {
	    if {[catch {Satisfiers $requirement} pkgList] \
		    || [llength $pkgList] == 0} {
		set Alternatives($requirementList) [list 1 "can't\
			satisfy requirement $requirement:\n  $pkgList"]
		break
	    }
	    # $pkgList contains packages which satisfy $requirement
	    # Combine these with their prerequisites to get a list of
	    # alternatives which satisfy $requirement.  Store that 
	    # list in $reqAlts.
	    set reqAlts [list]
	    set errors [list]
	    foreach {package pkgReq} $pkgList {

		if {![llength $pkgReq]} {
		    # No requirements in $pkgReq means $package is
		    # already provided, satisfying $requirement.
		    break
		}

		# Recursively get the alternatives which satisfy the
		# prerequisites for $package
		if {[catch {
			Alternatives $Prerequisites($package)
			} pkgAlts]} {
		    # An error finding alternative ways to satisfy the
		    # prerequisites for $pacakge.
		    # Record the error message.
		    set msg [join [split $pkgAlts \n] "\n    "]
		    lappend errors $package $msg
		    continue
		}

		# Each alternative way to satisfy the prerequisites,
		# extended by the requirement $pkgReq to load $package,
		# is an alternative which satisfys $requirement
		foreach alt $pkgAlts {
		    if {[catch {
			    ExtendAlternative $alt $package $pkgReq
			} extAlt]} {
			lappend errors $package $extAlt
		    } else {
			lappend reqAlts $extAlt
		    }
		}

	    }

	    if {![llength $reqAlts] && [llength $errors]} {
		# No alternatives satisfy $requirement
		set msg "can't satisfy requirement $requirement:"
		foreach {p m} $errors {
		    append msg "\n  $p: prerequisites:\
			    $Prerequisites($p)\n    $m"
		}
		set Alternatives($requirementList) [list 1 $msg]
		break
	    }

	    # $reqAlts is list of alternatives for satisyfing
	    # $requirement.  if $reqAlts is empty, $requirement is
	    # already satisfied.
	    if {![llength $reqAlts]} {
		lappend prior $requirement
		continue
	    }

	    # Merge each alternative from $reqAlts with the alternatives
	    # for satisfying all previous requirements in
	    # $requirementList
	    set newAlts [list]
	    set errors [list]
	    foreach oldAlt [lrange $Alternatives($requirementList) 1 end] {
		foreach reqAlt $reqAlts {
		    if {[catch {MergeAlternatives $oldAlt $reqAlt} newAlt]} {
			lappend errors $newAlt
		    } else {
			lappend newAlts $newAlt
		    }
		}
	    }

	    if {[llength $newAlts]} {
		set Alternatives($requirementList) \
			[linsert $newAlts 0 -1]
	    } else {
		set msg "can't satisfy requirement $requirement:\nall\
			alternatives inconsistent with prior\
			requirements: $prior"
		set i 0
		foreach e $errors {
		    append msg "\n  alternative [incr i]: $e"
		}
		set Alternatives($requirementList) [list 1 $msg]
		break
	    }

	    lappend prior $requirement

	}

	if {[lindex $Alternatives($requirementList) 0] == -1} {
	    set Alternatives($requirementList) \
		    [lreplace $Alternatives($requirementList) 0 0 0]
	}

    }
    switch -- [lindex $Alternatives($requirementList) 0] {
	0 {
	    return [lrange $Alternatives($requirementList) 1 end]
	}
	1 {
	    return -code error \
		    [lindex $Alternatives($requirementList) 1]
	}
	-1 {
	    return -code error "package dependency cycle:\n \
		    Requirement \"$requirementList\" eventually\
		    requires itself"
	}
    }
}

#		*	*	*	*	*

# [Satisfiers] accepts one requirement and returns a list which
# interleaves a list of all packages known to the interpreter that
# satisfy that requirement with the corresponding list of requirements
# which will cause each package to be loaded, in order of preference.
# The returned list may be empty.
#
# If the requirement conflicts with a provided package, an error is
# raised and an error message is returned.
#
# The Satisfiers array caches the results of this proc.  The value of
# Satisfiers($requirement) is a pair.  The first element is the return
# code of [Satisfiers $requirement].  The second element is the result
# of [Satisfiers $requirement].
#
proc ::package::Satisfiers {requirement} {

    # Get cached result, if any
    variable Satisfiers
    if {[info exists Satisfiers($requirement)]} {
	foreach {code result} $Satisfiers($requirement) break
	return -code $code $result
    }

    set package [lindex $requirement 0]
    if {![catch {eval package present $requirement} result]} {

	# The requirement is satisfied by a package already provided.
	# Return the package, but no requirement for further
	# prerequisites processing.
	#
	set Satisfiers($requirement) [list 0 \
		[list [list $package $result] [list]] ]
	return  [list [list $package $result] [list]]

    }
    if {![string match "" [package provide $package]]} {
	# A release of the required package is already provided that
	# does not satisfy the requirement.
	set Satisfiers($requirement) [list 1 $result]
	return -code error $result
    }

    # Construct list of versions of required package known to the interp
    set known [package versions $package]
    variable Prerequisites
    foreach pair [array names Prerequisites [list $package *]] {
	lappend known [lindex $pair 1]
    }

    # Sort, filter, and normalize the list of versions
    set vReq [lrange $requirement 1 end]
    set answer [list]
    foreach v [lsort -decreasing -command [list package compare] \
	    $known] {

        # Throw away non-satisfiers
        if {![eval package satisfies $v $vReq]} {continue}

	# Normalize version numbers
	while {[regsub {[.]0+$} $v {} v]} {}

	if {![llength $answer]} {
	    # First in list
	    lappend answer [list $package $v] $requirement

	} elseif {[package compare $v $vLast]} {
            # Appending to list: remove duplicates

	    lappend answer [list $package $v]
	    switch [llength $requirement] {
		1 {
		    lappend answer [list $package 0 $vLast]
		}
		2 {
		    lappend answer [linsert $requirement 2 $vLast]
		}
		3 {
		    lappend answer [lreplace $requirement 2 2 $vLast]
		}
	    }
	}
	set vLast $v
    }
    set Satisfiers($requirement) [list 0 $answer]
    set answer
}

#		*	*	*	*	*

# [ExtendAlternative] takes an alternative, a (package, version) pair,
# and the requirement corresponding to that (package, version) pair, and
# returns an alternative.  The returned alternative is the one passed
# in, extended by the package and requirement passed in.  The intent is
# that $pv is a package which has all of $alt as its prerequisites, and
# the returned alternative captures that relationship.
#
proc ::package::ExtendAlternative {alt pv req} {

    # Unpack map from alternative data structure
    foreach {DAG MAP} $alt {break}
    array set map $MAP

    set pkg [lindex $pv 0]
    if {[info exists map($pkg)]} {
	return -code error "package dependency cycle:\n  package\
		    \"$pv\" depends on requirement \"[lindex \
		    $map($pkg) 1]\""
    }
    set map($pkg) [list [lindex $pv 1] [lrange $req 1 end]]

    # Unpack dag and max from alternative data structure
    array set dag [lindex $DAG 0]
    set max [lindex $DAG 1]

    foreach vertex $max {
	lappend dag($vertex) $pkg
    }
    set dag($pkg) [list]
    set max [list $pkg]

    # return the extended alternative as a reassembled data structure
    list [list [array get dag] $max] [array get map]
}

#		*	*	*	*	*

# [MergeAlternatives] takes two alternatives and merges them into a
# single alternative which it returns.  The idea is that $alt1 is an
# alternative way to satisfy $requirementList1 and $alt2 is an
# alternative way to satisfy $requirementList2.  Then the alternative
# which is returned is an alternative to satisfy
# [concat $requirementList1 $requirementList2].
#
# If it is found that the two alternatives cannot be merged (they are
# inconsistent with one another), then an error is raised and an error
# message is returned.
#
proc ::package::MergeAlternatives {alt1 alt2} {

    # Unpack the maps from the alternative data structures
    foreach {DAG1 MAP1} $alt1 {break}
    foreach {DAG2 MAP2} $alt2 {break}
    array set map1 $MAP1
    array set map2 $MAP2
    if {[llength $MAP1] < [llength $MAP2]} {
	upvar 0 map1 smap map2 lmap
    } else {
	upvar 0 map2 smap map1 lmap
    }

    # First merge the maps.  If any package is provided by both $alt1
    # and $alt2, the versions provided must be the same.  If not,
    # raise an error.
    #
    foreach pkg [array names smap] {
	if {[info exists lmap($pkg)]} {
	    if {[package compare [lindex $smap($pkg) 0] \
		    [lindex $lmap($pkg) 0]]} {
		return -code error "conflicting versions of $pkg:\
		    [lindex $smap($pkg) 0] and [lindex $lmap($pkg) 0]"
	    }

	    # Versions match, so merge requirements.
	    set lmap($pkg) [lreplace $lmap($pkg) 1 1 \
		    [MergeRequirements [lindex $smap($pkg) 1] \
		    [lindex $lmap($pkg) 1]]]
	} else {
	    set lmap($pkg) $smap($pkg)
	}
    }
    # lmap now holds the merged map.

    # unpack the DAGs from the alternative data structures
    array set dag1 [lindex $DAG1 0]
    array set dag2 [lindex $DAG2 0]
    set max1 [lindex $DAG1 1]
    set max2 [lindex $DAG2 1]
    if {[llength $MAP1] < [llength $MAP2]} {
	upvar 0 dag1 sdag dag2 ldag max1 smax max2 lmax
    } else {
	upvar 0 dag2 sdag dag1 ldag max1 lmax max2 smax
    }

    # First be sure all maximal packages in either
    # DAG appear in both DAGs, possibly disconnected.
    foreach pkg $smax {
	lappend ldag($pkg)
    }
    foreach pkg $lmax {
	lappend sdag($pkg)
    }
    
    # Now merge the DAGs.
    set max [list]
    foreach pkg [array names sdag] {
	if {[info exists ldag($pkg)]} {
	    set ldag($pkg) [Union $sdag($pkg) $ldag($pkg)]
	    if {![llength $ldag($pkg)]} {
		lappend max $pkg
	    }
	} else {
	    set ldag($pkg) $sdag($pkg)
	}
    }

    # return the merged alternative as a reassembled data structure
    list [list [array get ldag] $max] [array get lmap]
}

# [TopologicalSort] takes an alternative (which is represented by a DAG)
# and returns a topological sort of it.  A topological sort of a DAG is
# a list of all the vertices of the DAG (here a list of package names),
# such that if A -> B is in the DAG, then the index of A in the list is
# less than the index of B in the list.  In addition, when constructing
# the list, this proc includes the corresponding requirement along with
# each package name.  Thus, the result is a requirement list, ordered so
# that each requirement is listed after its prerequisites.
#
proc ::package::TopologicalSort {alt} {

    # Unpack the alternative data structure
    foreach {DAG MAP} $alt {break}
    array set dag [lindex $DAG 0]
    set max [lindex $DAG 1]
    array set map $MAP

    set answer [list]
    foreach vertex $max {
	unset dag($vertex)
    }
    while {[llength $max]} {
	set pkg [lindex $max 0]
	set max [lrange $max 1 end]
	set answer [linsert $answer 0 \
		[linsert [lindex $map($pkg) 1] 0 $pkg]]
	foreach vertex [array names dag] {
	    set idx [lsearch -exact $dag($vertex) $pkg]
	    set dag($vertex) [lreplace $dag($vertex) $idx $idx]
	    if {[llength $dag($vertex)] == 0} {
		lappend max $vertex
		unset dag($vertex)
	    }
	}
    }
    set answer
}

# [MergeRequirements] takes two requirements (without the package name)
# and returns the intersection of those two requirements.
#
# Since this routine will only be called when it is already known that
# the two requirements yield the same version, there's no need to check
# for a conflicting pair of requirements.
#
proc ::package::MergeRequirements {r1 r2} {
    if {[llength $r2] == 0} {
        return $r1
    }
    if {[llength $r1] == 0} {
        return $r2
    }
    if {[llength $r1] == 1} {
        regexp {^[0-9]+} [lindex $r1 0] major
        lappend r1 [incr major]
    }
    if {[llength $r2] == 1} {
        regexp {^[0-9]+} [lindex $r2 0] major
        lappend r2 [incr major]
    }

    foreach {min1 max1} $r1 {break}
    foreach {min2 max2} $r2 {break}
    if {[package compare $min1 $min2] < 0} {
        set min $min2
    } else {
        set min $min1
    }
    if {[package compare $max1 $max2] < 0} {
        set max $max1
    } else {
        set max $max2
    }
    list $min $max
}

# [Union] takes two lists and returns their union as sets. The elements
# of these lists are package names.  We must use argument names which
# will not appear in the lists.  Since a package name may be any string
# at all, we can't acheive that goal with certainty.  Best we can do is
# pick argument names which seem very unlikely to be used as package
# names.
#
proc ::package::Union [list \1 \1\1] {
    foreach [lappend \1 \1] {.} [lappend \1\1 \1\1] {} {}
    unset \1 \1\1
    info locals
}

