# FILE: package.tcl
#
# Initialization file for the package "package"
#
# The package "package" supplies an upward compatible replacement for
# the [package] command in the Tcl core which supplies better support
# for managing conflicts in package dependencies.

# This package uses namespaces
package require Tcl 8

namespace eval ::package {

    # Public variables and exported commands
    variable version 0.0.1
    variable library [file join [pwd] [file dirname [info script]]]
    namespace export package

}

# Utility routine for wrapping calls to the renamed core [package] or
# other support procs in this package so this mechanism is not apparent
# to the caller.
proc ::package::Wrap {rename map {len 0}} {
    return [format {
	set cmd [lreplace [info level 0] 0 %d %s]
	if {[set code [catch {uplevel 1 $cmd} msg]] == 1} {
	    set errList [split $::errorInfo \n]
	    set errList [lrange $errList 0 [expr {[llength $errList] \
		    - [llength [split $cmd \n]] - 5}]]
	    set newErrorInfo [join $errList \n]
	    foreach var {msg newErrorInfo} {
		regsub -all {%s} [set $var] \
			[lrange [info level 0] 0 %d] $var
		foreach {pre post} {%s} {
		    regsub -all $pre [set $var] $post $var
		}
	    }
	    return -code error -errorinfo $newErrorInfo $msg
	}
	return -code $code $msg} $len $rename $rename $len $map]
}

# This is the replacement [package] provided by this package.
# It adds the subcommands "compare", "index", "mkdepend",
# "prerequisites", and "satisfies" and extends subcommands
# "present", "require", and "vsatisfies".
#
proc ::package::package {args} {
    set option [lindex $args 0]
    # Substitutions to make in error messages:
    set errorMap {
	"names, provide"	"names, present, provide"
	"names, present"	"names, prerequisites, present"
	"ifneeded, names"	"ifneeded, index, mkdepend, names"
	"require, unknown"	"require, satisfies, unknown"
	"vcompare, versions, or vsatisfies"	"or versions"
	"be forget"		"be compare, forget"
    }
    switch -glob -- $option {
	c* {
	    if {[string first $option compare] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::compare $errorMap 1]
	}
	in* {
	    if {[string first $option index] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::index $errorMap 1]
	}
	m* {
	    if {[string first $option mkdepend] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::mkdepend $errorMap 1]
	}
	prer* {
	    if {[string first $option prerequisites] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::prerequisites $errorMap 1]
	}
	pres* {
	    if {[string first $option present] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::present $errorMap 1]
	}
	r* {
	    if {[string first $option require] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::require $errorMap 1]
	}
	s* {
	    if {[string first $option satisfies] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::satisfies $errorMap 1]
	}
	vs* {
	    if {[string first $option vsatisfies] != 0} {
		eval [Wrap ::package:: $errorMap]
	    }
	    eval [Wrap ::package::satisfies $errorMap 1]
	}
	default {
	    eval [Wrap ::package:: $errorMap]
	}
    }
}

# Alias [package compare] to [package vcompare]
interp alias {} ::package::compare {} package vcompare

# Normally a package should not import its commands into the global
# namespace.  Here an exception to that rule is necessary, because the
# [package] command must manage the interpreter-wide collection of
# packages.  Thus, it is critically important that all namespaces use
# the same [package] command.  This is achieved by putting the
# [package] command in the global namespace.
rename package ::package::
namespace import ::package::package

# Set up for auto-loading the subcommands.
if {[lsearch -exact $auto_path $::package::library] == -1} {
    lappend auto_path $::package::library
}

# Everything is ready; provide the package
package provide package $::package::version

# Now that the package is provided, re-[source] the pkgIndex.tcl files
# so that the ones which want to use the interface this package
# provides will be able to do so.
eval [package unknown] Tcl [package provide Tcl]

