# FILE: prerequisites.tcl
#
# The [package prerequisites] command of the package "package".
#
# Usage:
# 	package prerequisites $package $version ?$prerequisiteList?
#
# Register that version $version of package $package depends on the
# interface prerequisites in $prerequisiteList.  Each element of
# $prerequisiteList must be a list of arguments suitable for passing to
# [package require].  If $prerequisiteList is omitted, return the
# current list of prerequisites registered for version $version of
# package $package.
#
proc ::package::prerequisites {package version args} {

    # Check for valid version number
    if {[catch {package compare $version $version} msg]} {
	return -code error $msg
    }

    # Remove trailing ".0" in version numbers so that each version
    # maps to a unique slot in the Prerequisites array.
    while {[regsub {[.]0+$} $version {} version]} {}

    variable Prerequisites
    set pair [list $package $version]
    if {[llength $args] == 0} {
	if {[info exists Prerequisites($pair)]} {
	    return $Prerequisites($pair)
	} else {
	    return [list]
	}
    }
    if {[llength $args] != 1} {
	return -code error "wrong # args: should be \"[lindex \
		[info level 0] 0] package version ?prerequisiteList?\""
    }

    # Check that $prerequisiteList is valid Tcl list
    set prerequisiteList [lindex $args 0]
    if {[catch {llength $prerequisiteList} msg]} {
	return -code error "bad prerequisiteList: $msg"
    }

    # Register the list of package interface prerequisites
    set Prerequisites($pair) $prerequisiteList
}

