# FILE: require.tcl
#
# The [package require] command of the package "package".
#
# Usage:
# 	package require ?-exact? $package ?$minVersion? ?$maxVersion?
#
# The [package require] command is meant to be a compatible extended
# version of the [package require] command in the Tcl core.  
#
# It behaves the same as the the [package require] command in the Tcl
# core, except that it uses [package satisfies] to compare the version
# number of available releases of $package against the $minVersion and
# $maxVersion arguments to determine whether an available package
# satisfies the requirement.
#
proc ::package::require {args} {
  
    # Pass along to the core "package require" any $args which it
    # can handle.
    if {[string match -exact [lindex $args 0]] 
	    || [llength $args] != 3} {
	eval [Wrap "::package:: require" ""]
    }

    # Okay, $args is [list $package $minVersion $maxVersion]
    set package [lindex $args 0]

    set provided [package provide $package]
    if {[string length $provided]} {
	if {[catch {
		eval [lreplace $args 0 0 package satisfies $provided]
		} msg]} {
	    return -code error $msg
	}
	if {!$msg} {
	    return -code error "version conflict for package\
		    \"$package\": have $provided, need\
		    [lrange $args 1 end]"
	}
	return $provided
    }

    set avail [FilterByRequirement [package versions $package] $args]

    if {[llength $avail] == 0} {
	set cmd [package unknown]

	# Evaluate [package unknown] handler in global scope
	# NOTE: This passes all arguments of [package require] on
	# to the [package unknown] handler.  This may break
	# compatibility with customized [package unknown] handlers
	# due to extra arguments.

	if {[catch {uplevel #0 $cmd $args} msg]} {
	    global errorInfo errorCode
	    return -code error -errorinfo \
		    "$errorInfo\n    (\"package unknown\" script)" \
		    -errorcode $errorCode $msg
	}

        set avail [FilterByRequirement \
		[package versions $package] $args]
    }

    if {[llength $avail] == 0} {
	return -code error "can't find package satisfying $args"
    }

    set cmd [list ::package:: require -exact]
    set code [catch {uplevel 1 $cmd [lindex $avail 0]} msg]
    return -code $code $msg
}

# Given a list of version numbers and a requirement, returns a list of
# packages with version numbers drawn from the input list.  All packages
# in the returned list satisfy the requirement, and are in decreasing
# order of preference, without duplicates.
#
proc ::package::FilterByRequirement {vList requirement} {
    set package [lindex $requirement 0]
    set vReq [lrange $requirement 1 end]
    set answer [list]
    foreach v [lsort -decreasing -command [list package compare] \
	    $vList] {

	# Throw away non-satisfiers
	if {![eval package satisfies $v $vReq]} {continue}

	# Remove duplicates
	if {![llength $answer] || [package compare $v \
		[lindex [lindex $answer end] 1]]} {
	    lappend answer [list $package $v]
	}
    }
    set answer
}

