# FILE: satsifies.tcl
#
# The [package satsifes] command of the package "package".
#
# Usage:
# 	package satisfies $version ?$minVersion? ?$maxVersion?
#
# The [package satisfies] command is meant to be a compatible extended
# version of the [package vsatisfies] command.  
#
# The question arises in package management whether package
# "$package $version" will satisfy a requirement
# "$neededPackage ?$minVersion? ?$maxVersion?" made known to
# [package require].  [package satisfies] answers half of that question.
# The requirement is satisfied iff $package and $neededPackage are
# identical strings and
# [package satisfies $version ?$minVersion? ?$maxVersion?] returns true.
#
# If both $minVersion and $maxVersion are omitted, [package satisfies]
# will always return true.  So, package "foo 2.4" satisfies requirement
# "foo".
#
# If both $minVersion and $maxVersion are supplied and are equal,
# [package satisfies] returns true iff $version == $minVersion.  So,
# package "foo 2.4" satisfies requirement "foo 2.4 2.4".  This case
# is meant to replace the "-exact" switch of [package require].
#
# If both $minVersion and $maxVersion are supplied and are not equal,
# [package satisfies] returns true iff $version >= $minVersion and
# $version < $maxVersion.  So, package "foo 2.4" satisfies requirement
# "foo 2.1 2.7" and it also satisfies requirement "foo 1.4 5.0".
#
# If only $minVersion is supplied, its major version number is extracted
# and incremented and the result is taken to be the value of $maxVersion
# in determining satisfaction.  Thus package "foo 2.4" satisfies
# requirement "foo 2.4" but does not satisfy requirement "foo 1.5".
#
# This subcommand is renamed from "vsatisfies" to "satisfies" in the
# interest of making the overall API of the replacement [package]
# command less cryptic and more like simple English words and phrases.
# It is analogous to the move away from "lreplace" to "list replace".
# The old subcommand is still supported as an alias to this one for
# compatibility.
#
proc ::package::satisfies {version args} {

    # Check for valid version number
    if {[catch {package compare $version $version} msg]} {
	return -code error $msg
    }
    if {[llength $args] > 2} {
	return -code error "wrong # args: should be \"package satisifes\
		version ?minVersion? ?maxVersion?\""
    }
    if {[llength $args] == 0} {
	return 1
    }
    foreach v $args {
	if {[catch {package compare $v $v} msg]} {
	    return -code error $msg
	}
    }
    set minVersion [lindex $args 0]
    if {[package compare $version $minVersion] < 0} {
	return 0
    }
    if {[llength $args] == 2} {
	set maxVersion [lindex $args 1]
	if {![package compare $minVersion $maxVersion]} {
	    return [expr {![package compare $version $minVersion]}]
	}
    } else {
	regexp {^[0-9]+} $minVersion minMajor
	set maxVersion [incr minMajor]
    }
    if {[package compare $version $maxVersion] >= 0} {
	return 0
    }
    return 1
}

